//=======================UART_FullDuplex=====================
//Engineer: Jose Pablo Pinilla
//Engineer: Jose Manuel Velandia
//Website:	semilleroadt.upbbga.edu.co


module UART_FullDuplex (
	input 		clk,
	//Transmitter
	input 		Tx_Start,
	input 		[7:0]Tx_Data,
	output 		Tx,
	output		TxD_Busy,
	//Receiver
	input 		Rx,
	output reg	[7:0]Rx_Data,
	output		RxD_Ready
	);
	

//========================Tx/Rx Control=====================
reg		Frame_Rx=0;						//Frame is being received
reg		Frame_Tx=0;						//Frame is being transmitted
assign	TxD_Busy = Frame_Tx;
assign 	RxD_Ready= ~Frame_Rx;
//========================Tx/Rx Control=====================


parameter clk_in_freq = 50000000;
parameter baud_rate = 9600;
localparam counter_limit = clk_in_freq/(2*baud_rate);


//=====================Baud Rate Generation Tx==================
reg [15:0]div_clk_counter_tx=0;
reg baud_clk_tx=0;
wire	clk_count_reset_tx = (div_clk_counter_tx>=(counter_limit-1));//Change;

always @ (posedge clk or posedge clk_count_reset_tx)
	if (clk_count_reset_tx)
		div_clk_counter_tx <= 16'd0;
	else
		begin
			if (Frame_Tx)
				div_clk_counter_tx <= div_clk_counter_tx + 1;
			else
				div_clk_counter_tx <= 16'd0;//Change
		end		

always @ (posedge clk_count_reset_tx)
	baud_clk_tx <= ~baud_clk_tx;
//=====================Baud Rate Generation Tx==================



//=====================Baud Rate Generation Rx==================
reg [15:0]div_clk_counter_rx=0;
reg baud_clk_rx=0;
wire clk_count_reset_rx	=	(div_clk_counter_rx>=(counter_limit-1));

always @ (posedge clk or posedge clk_count_reset_rx)
	if (clk_count_reset_rx)
		div_clk_counter_rx <= 16'd0;
	else
	begin
		if (Frame_Rx)
			div_clk_counter_rx <= div_clk_counter_rx + 1;
		else
			div_clk_counter_rx <= 16'd0;
	end

always @ (posedge clk_count_reset_rx)
	baud_clk_rx <= ~baud_clk_rx;
//=====================Baud Rate Generation Rx==================



//======================Tx========================
reg [3:0]Tx_counter=0;
reg [9:0]OutBuffer = 10'b11111111_0_1;
wire Tx_Counter_Reset = (Tx_counter>=10);//Change '>=' Means avoid troubles '==' can be dangerous if something wrong happens

//Tx Data Frame Counter
always @ (negedge baud_clk_tx or negedge Frame_Tx)
	if (~Frame_Tx)
		Tx_counter <= 0;
	else
		Tx_counter <= Tx_counter+4'd1;

//Tx Detection
always@(posedge Tx_Start or posedge Tx_Counter_Reset)
	if (Tx_Counter_Reset)
		Frame_Tx <= 0;
	else
		Frame_Tx <= 1;

//Tx Shift Register 
always @ (posedge baud_clk_tx or	negedge Frame_Tx)
	if (~Frame_Tx)
		OutBuffer <= {Tx_Data,2'b01};	///Asynchronous Load // Always loads 'Tx_Data' inside 'OutBuffer' when 'Frame_Tx' has low level (0)
	else
		OutBuffer <= {1'b1,OutBuffer[9:1]};
//======================Tx========================


	
//======================Rx=========================
reg [8:0]InBuffer=0;//Change .. Rx_Buffer doesn't need more than 9 bits for recognition (1 bit start + 8 bits info)
reg [3:0]Rx_counter=0;
wire Rx_Counter_Reset = (Rx_counter>=4'd9);//Change '>=' Means avoid troubles '==' can be dangerous if something wrong happens

//Rx Data Frame Counter	
always @ (negedge baud_clk_rx or negedge Frame_Rx)
	if (~Frame_Rx)
		Rx_counter <= 4'd0;
	else
		Rx_counter <= Rx_counter	+	4'd1;

//Rx Detection
always@(negedge  Rx or posedge Rx_Counter_Reset)
	if (Rx_Counter_Reset)
	begin
		Frame_Rx <= 0;
		Rx_Data <= InBuffer[8:1];//Change .. Before scaning the last 8 info bit, the reception is finishing and 'Rx' input stay alert for 'negedge' signal
	end
	else
		Frame_Rx <= 1;
	
//Rx Shift Register
always @ (posedge baud_clk_rx)
	if (Frame_Rx)
		InBuffer[8:0] <= {Rx,InBuffer[8:1]};//Change .. The last 8 info bit (1 or 0) and the stop bit (1) can provide a 'posedge' or a simple 'high level' on the 'Rx' signal input; this situation doesn't create a false alarm for reception (negedge signal) on this input
//======================Rx=========================


		
//================Output Assignment==============
assign Tx = OutBuffer[0];
//assign Rx_Data = InBuffer[8:1];//Rx_Data is registered on line #124 so the system can read this out even if it gets a little delayed
//================Output Assignment==============

endmodule 